from cmflib import cmfquery
import pandas as pd
import itertools
from ml_metadata.proto.metadata_store_pb2 import Value

async def query_exec_lineage(mlmd_path, pipeline_name, dict_of_exe_ids, exec_type, uuid_server):
    data = {}
    query = cmfquery.CmfQuery(mlmd_path)
    pipeline_id = query.get_pipeline_id(pipeline_name)
    node_id_name_list = []
    link_src_trgt_list = []
    host_id = None
    for id, context_type, uuid in zip(dict_of_exe_ids[pipeline_name]["id"], dict_of_exe_ids[pipeline_name]["Context_Type"], dict_of_exe_ids[pipeline_name]["Execution_uuid"]):
        truncated_uuid = uuid.split("_")[0][:4]
        # this condition is failing that's why we are unable to assign id to host_id
        if (context_type + "_" + truncated_uuid) == (pipeline_name + "/" + exec_type + "_" + uuid_server):
            host_id = id
            node_id_name_list.append({"id":host_id, "name":(context_type.split("/")[-1] + "_" + uuid.split("-")[0][:4]), "color":"#16B8E9"})
    #exec=query.get_one_hop_parent_executions([host_id],pipeline_id)
    exec_new = None
    if host_id is None:
        return data
    exec_new = query.get_all_parent_executions_by_id([host_id], pipeline_id)
    if exec_new is None:
        return data
    for i in exec_new[0]:
        id = i[0]
        name = i[1]
        exec_uuid = i[2]
        node_id_name_list.append({"id":i[0], "name":(name.split("/")[-1] + "_" + exec_uuid.split("-")[0][:4]), "color":"#FA6318"})
    link_src_trgt_list.append(exec_new[1])
    if not node_id_name_list and not link_src_trgt_list:
        return data
    node_id_name_list_unique = [dict(t) for t in {tuple(d.items()) for d in node_id_name_list}]
    new_list = pd.DataFrame(
      link_src_trgt_list[0]
    ).drop_duplicates().to_dict('records')
    data = {
        "nodes" : node_id_name_list_unique,
        "links" : new_list
    }
    return data

#query_exec_lineage("/home/chobey/cmf-server/data/mlmd", "Test-env",data,"Evaluate")



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import boto3

class AmazonS3Artifacts:
    def download_artifacts(
        self,
        dvc_config_op,
        current_directory: str,
        bucket_name: str,
        object_name: str,
        download_loc: str,
    ):
        access_key = dvc_config_op["remote.amazons3.access_key_id"]
        secret_key = dvc_config_op["remote.amazons3.secret_access_key"]
        session_token = dvc_config_op["remote.amazons3.session_token"]
        try:
            s3 = boto3.client(
                's3',
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_key,
                aws_session_token=session_token
            )
            s3.head_bucket(Bucket=bucket_name)
            dir_path = ""
            if "/" in download_loc:
                dir_path, _ = download_loc.rsplit("/", 1)
            if dir_path != "":
                os.makedirs(dir_path, mode=0o777, exist_ok=True)  # creating subfolders if needed
            response = s3.download_file(bucket_name, object_name, download_loc)
            if response == None:
                return f"{object_name} downloaded at {download_loc}"
            return response

        except s3.exceptions.ClientError as e:
            # If a specific error code is returned, the bucket does not exist
            if e.response['Error']['Code'] == '404':
                return f"{bucket_name}  doesn't exists!!"
            else:
                # Handle other errors
               raise
        except TypeError as exception:
            return exception
        except Exception as e:
            return e



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os

from dvc.api import DVCFileSystem

class LocalArtifacts:
    def download_artifacts(
        self,
        dvc_config_op,
        current_directory: str,
        current_loc: str,
        download_loc: str,
    ):
        obj = True
        try:
            fs = DVCFileSystem(
                dvc_config_op["remote.local-storage.url"]
            )  # dvc_config_op[1] is file system path - "/path/to/local/repository"
            # get_file() only creates file, to put artifacts in proper directory, subfolders are required.
            temp = download_loc.split("/")
            temp.pop()
            dir_path = "/".join(temp)
            os.makedirs(dir_path, mode=0o777, exist_ok=True)  # creating subfolders
            obj = fs.get_file(current_loc, download_loc)
            if obj == None:  # get_file() returns none when file gets downloaded.
                stmt = f"object {current_loc} downloaded at {download_loc}."
                return stmt
        except TypeError as exception:
            return exception
        except Exception as exception:
            return exception



from cmflib import cmfquery, cmf_merger
import pandas as pd
import json,glob
import os
from server.app.query_visualization import query_visualization
from server.app.query_visualization_execution import query_visualization_execution
from fastapi.responses import FileResponse

async def get_executions_by_ids(mlmdfilepath, pipeline_name, exe_ids):
    '''
    Args:
     mlmdfilepath: mlmd file path.
     pipeline_name: name of the pipeline.
     exe_ids: list of execution ids.
   
    Returns:
     returns dataframe of executions using execution_ids.
    '''
    query = cmfquery.CmfQuery(mlmdfilepath)
    df = pd.DataFrame()
    executions = query.get_all_executions_by_ids_list(exe_ids)
    df = pd.concat([df, executions], sort=True, ignore_index=True)
    #df=df.drop('name',axis=1)
    return df

async def get_all_exe_ids(mlmdfilepath):
    '''
    Returns:
        returns a dictionary which has pipeline_name as key and dataframe which includes {id,Execution_uuid,Context_Type,Context_id} as value.
    '''
    query = cmfquery.CmfQuery(mlmdfilepath)
    execution_ids = {}
    names = query.get_pipeline_names()
    for name in names:
        df = pd.DataFrame()    # df is emptied to store execution ids for next pipeline.
        stages = query.get_pipeline_stages(name)
        for stage in stages:
            executions = query.get_all_executions_in_stage(stage)
            df = pd.concat([df, executions], sort=True, ignore_index=True)
        # check if df is empty return just pipeline_name: {}
        # if df is not empty return dictionary with pipeline_name as key 
        # and df with id, context_type, uuid, context_ID as value.
        if not df.empty:
            execution_ids[name] = df[['id', 'Context_Type', 'Execution_uuid', 'Context_ID']]
        else:
            execution_ids[name] = pd.DataFrame()
    return execution_ids

async def get_all_artifact_ids(mlmdfilepath):
    # following is a dictionary of dictionary
    # First level dictionary key is pipeline_name
    # First level dicitonary value is nested dictionary
    # Nested dictionary key is type i.e. Dataset, Model, etc.
    # Nested dictionary value is ids i.e. set of integers
    artifact_ids = {}
    query = cmfquery.CmfQuery(mlmdfilepath)
    names = query.get_pipeline_names()
    execution_ids = await get_all_exe_ids(mlmdfilepath)
    for name in names:
        df = pd.DataFrame()
        if not execution_ids.get(name).empty:
            exe_ids = execution_ids[name]['id'].tolist()
            for id in exe_ids:
                artifacts = query.get_all_artifacts_for_execution(id)
                df = pd.concat([df, artifacts], sort=True, ignore_index=True)
            #acknowledging pipeline exist even if df is empty. 
            if df.empty:
                artifact_ids[name] = pd.DataFrame()   # { pipeline_name: {empty df} }
            else:
                df.sort_values("id", inplace=True)
                df.drop_duplicates(subset="id", keep='first', inplace=True)
                artifact_ids[name] = {}
                for art_type in df['type']:
                    filtered_values = df.loc[df['type'] == art_type, ['id', 'name']]
                    artifact_ids[name][art_type] = filtered_values
        # if execution_ids is empty create dictionary with key as pipeline name
        # and value as empty df
        else:
            artifact_ids[name] = pd.DataFrame()
    return artifact_ids

async def get_artifacts(mlmdfilepath, pipeline_name, art_type, artifact_ids):
    query = cmfquery.CmfQuery(mlmdfilepath)
    names = query.get_pipeline_names()  # getting all pipeline names in mlmd
    df = pd.DataFrame()
    for name in names:
        if name == pipeline_name:
            df = query.get_all_artifacts_by_ids_list(artifact_ids)
            if len(df) == 0:
                return
            df = df.drop_duplicates()
            art_names = df['name'].tolist()
            name_dict = {}
            name_list = []
            exec_type_name_list = []
            exe_type_name = pd.DataFrame()
            for name in art_names:
                executions = query.get_all_executions_for_artifact(name)
                exe_type_name = pd.concat([exe_type_name, executions], ignore_index=True)
                execution_type_name = exe_type_name["execution_type_name"].drop_duplicates().tolist()
                execution_type_name = [str(element).split('"')[1] for element in execution_type_name]
                execution_type_name_str = ',\n '.join(map(str, execution_type_name))
                name_list.append(name)
                exec_type_name_list.append(execution_type_name_str)
            name_dict['name'] = name_list
            name_dict['execution_type_name'] = exec_type_name_list
            name_df = pd.DataFrame(name_dict)
            merged_df = df.merge(name_df, on='name', how='left')
            merged_df['name'] = merged_df['name'].apply(lambda x: x.split(':')[0] if ':' in x else x)
            merged_df = merged_df.loc[merged_df["type"] == art_type]
            result = merged_df.to_json(orient="records")
            tempout = json.loads(result)
            return tempout

def get_artifact_types(mlmdfilepath):
    query = cmfquery.CmfQuery(mlmdfilepath)
    artifact_types = query.get_all_artifact_types()
    return artifact_types

async def create_unique_executions(server_store_path, req_info):
    mlmd_data = json.loads(req_info["json_payload"])
    pipelines = mlmd_data["Pipeline"]
    pipeline = pipelines[0]
    pipeline_name = pipeline["name"]
    executions_server = []
    list_executions_exists = []
    if os.path.exists(server_store_path):
        query = cmfquery.CmfQuery(server_store_path)
        stages = query.get_pipeline_stages(pipeline_name)
        for stage in stages:
            executions = []
            executions = query.get_all_executions_in_stage(stage)
            for i in executions.index:
                for uuid in executions['Execution_uuid'][i].split(","):
                    executions_server.append(uuid)
        executions_client = []
        for i in mlmd_data['Pipeline'][0]["stages"]:  # checks if given execution_id present in mlmd
            for j in i["executions"]:
                if j['name'] != "": #If executions have name , they are reusable executions
                    continue       #which needs to be merged in irrespective of whether already
                                   #present or not so that new artifacts associated with it gets in.
                if 'Execution_uuid' in j['properties']:
                    for uuid in j['properties']['Execution_uuid'].split(","):
                        executions_client.append(uuid)
                else:
                    # mlmd push is failed here
                    status="version_update"
                    return status
        if executions_server != []:
            list_executions_exists = list(set(executions_client).intersection(set(executions_server)))
        for i in mlmd_data["Pipeline"]:
            for stage in i['stages']:
                for cmf_exec in stage['executions'][:]:
                    uuids = cmf_exec["properties"]["Execution_uuid"].split(",")
                    for uuid in uuids:
                        if uuid in list_executions_exists:
                            stage['executions'].remove(cmf_exec)

        for i in mlmd_data["Pipeline"]:
            i['stages']=[stage for stage in i['stages'] if stage['executions']!=[]]
    for i in mlmd_data["Pipeline"]:
        if len(i['stages']) == 0 :
            status="exists"
        else:
            cmf_merger.parse_json_to_mlmd(
                json.dumps(mlmd_data), "/cmf-server/data/mlmd", "push", req_info["id"]
            )
            status='success'
    return status


async def get_mlmd_from_server(server_store_path, pipeline_name, exec_id):
    query = cmfquery.CmfQuery(server_store_path)
    execution_flag = 0
    # checks if given execution_id present in mlmd
    if (
        pipeline_name in query.get_pipeline_names()
    ):  # checks if pipeline name is available in mlmd
        json_payload = query.dumptojson(pipeline_name, None)
        mlmd_data = json.loads(json_payload)["Pipeline"]
        if exec_id == None:
            json_payload = query.dumptojson(pipeline_name, exec_id)
        else:
            for stage in mlmd_data[0]["stages"]:
                for execution in stage["executions"]:
                    if execution["id"] == int(exec_id):
                        execution_flag = 1
                        break
            if execution_flag == 1:
                json_payload = query.dumptojson(pipeline_name, exec_id)
            else:
                json_payload = "no_exec_id"
                return json_payload
    else:
        json_payload = "NULL"
    return json_payload

async def get_lineage_data(server_store_path,pipeline_name,type,dict_of_art_ids,dict_of_exe_ids):
    query = cmfquery.CmfQuery(server_store_path)
    if type=="Artifacts":
        lineage_data = query_visualization(server_store_path, pipeline_name, dict_of_art_ids)
        '''
        returns dictionary of nodes and links for artifact lineage.
        lineage_data= {
                       nodes:[],
                       links:[]
                      }
        '''
    elif type=="Execution":
        lineage_data = query_visualization_execution(server_store_path, pipeline_name, dict_of_art_ids, dict_of_exe_ids)
        '''
        returns list of execution types for specific pipeline.
        '''
    else:
        lineage_data = query_visualization_ArtifactExecution(server_store_path, pipeline_name)
    return lineage_data




###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

#!/usr/bin/env python3
import argparse
import os
import subprocess

from cmflib.cli.command import CmdBase
from cmflib.cli.utils import check_minio_server
from cmflib.utils.dvc_config import DvcConfig
from cmflib.dvc_wrapper import dvc_push


class CmdArtifactPush(CmdBase):
    def run(self):
        result = ""
        dvc_config_op = DvcConfig.get_dvc_config()
        out_msg = check_minio_server(dvc_config_op)
        if dvc_config_op["core.remote"] == "minio" and out_msg != "SUCCESS":
            return out_msg
        else:
            result = dvc_push()
            return result


def add_parser(subparsers, parent_parser):
    HELP = "Push artifacts to the user configured artifact repo."

    parser = subparsers.add_parser(
        "push",
        parents=[parent_parser],
        description="Push artifacts to the user configured artifact repo.",
        help=HELP,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.set_defaults(func=CmdArtifactPush)



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
from minio import Minio
from minio.error import S3Error


class MinioArtifacts:
    def download_artifacts(
        self,
        dvc_config_op,
        current_directory: str,
        bucket_name: str,
        object_name: str,
        file_path: str,
    ):
        endpoint = dvc_config_op["remote.minio.endpointurl"].split("http://")[1]
        access_key = dvc_config_op["remote.minio.access_key_id"]
        secret_key = dvc_config_op["remote.minio.secret_access_key"]
        try:
            client = Minio(
                endpoint, access_key=access_key, secret_key=secret_key, secure=False
            )
            found = client.bucket_exists(bucket_name)
            if not found:
                return "Bucket doesn't exists"
            obj = client.fget_object(bucket_name, object_name, file_path)
            if obj:
                stmt = f"object {object_name} downloaded at {file_path}."
                return stmt
            else:
                return f"object {object_name} is not downloaded."

        except TypeError as exception:
            return exception
        except S3Error as exception:
            return exception



###
# Copyright (2023) Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

import os
import paramiko


class SSHremoteArtifacts:
    def download_artifacts(
        self,
        dvc_config_op,
        host: str,
        current_directory: str,
        remote_file_path: str,
        local_path: str,
    ):
        output = ""
        remote_repo = dvc_config_op["remote.ssh-storage.url"]
        user = dvc_config_op["remote.ssh-storage.user"]
        password = dvc_config_op["remote.ssh-storage.password"]
        try:
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(
                paramiko.AutoAddPolicy()
            )  # this can lead to man in the middle attack, need to find another solution
            ssh.connect(host, username=user, password=password)
            sftp = ssh.open_sftp()
            temp = local_path.split("/")
            temp.pop()
            dir_path = "/".join(temp)
            dir_to_create = os.path.join(current_directory, dir_path)
            os.makedirs(
                dir_to_create, mode=0o777, exist_ok=True
            )  # creates subfolders needed as per artifacts folder structure
            local_file_path = os.path.join(current_directory, local_path)
            local_file_path = os.path.abspath(local_file_path)
            output = sftp.put(remote_file_path, local_file_path)
            sftp.close()
            ssh.close()
            if output:
                stmt = f"object {remote_file_path} downloaded at {local_file_path}."
                return stmt

        except TypeError as exception:
            return exception
        except Exception as exception:
            return exception
